# :title: Rack::State -- Secure & Modular State Management
# :main:  README.rdoc

require 'securerandom'

module Rack

  #
  # == The Middleware Interface
  #
  # Stores and manages the state of a single object on the server and sets the
  # state token in a client cookie. Multiple State instances may be utilized to
  # manage multiple states, each with different storage adapters and options.
  #
  # State initializes Manager, which contains the API.
  #
  class State

    VERSION = '0.0.3'

    #
    # == Middleware Options
    #
    # [store]
    #   Instance of the storage adapter. See Rack::State::Store for available
    #   adapters. The default is Store::Memory.
    # [key]
    #   Client cookie name and Rack environment key suffix. For example, the key
    #   "token" sets a cookie named "token" and the environment key to
    #   "rack.state.token". Therefore, the application can access the
    #   State::Manager instance at <tt>env["rack.state.token"]</tt>.
    #   The default is "token".
    # [domain, path, max_age, expires, secure, httponly]
    #   Standard cookie options supported by Rack. None are set by default;
    #   therefore, the state will be valid for the session only on the domain
    #   and path in which it was set and available over unsecured connections.
    #
    def initialize(app, options = {})
      @app = app
      @options = options
      @store = options.delete(:store) || Store::Memory.new
      @key = options.delete(:key) || 'token'
      @skey = "rack.state.#{@key}"
    end

    def call(env)
      request = Request.new(env)

      if request.path =~ /^#{@options[:path]}/
        token = request.cookies.fetch(@key, nil)
        state = env[@skey] = Manager.new(@store, token, @key, @options)
        status, headers, body = @app.call(env)
        resp = Response.new(body, status, headers)
        state.finish(resp)
        resp.finish
      else
        @app.call(env)
      end
    end

    #
    # == The Application Interface
    #
    # Orchestrates the state management of an object. The API is very simple:
    # you can #get, #set, and #delete an object.
    #
    # The application can access Manager via the Rack environment.
    # See the *key* middleware option in State.new.
    #
    class Manager

      #
      # Instantiated in middleware-land.
      #
      # [store]
      #   Storage instance that implements CRUD methods for storing state.
      # [token]
      #   Used to reference the state object in store. The token is stored as 
      #   the cookie value on the client. If nil, token not set on client.
      # [key]
      #   Cookie name.
      # [options]
      #   Cookie options.
      #
      def initialize(store, token, key, options) #:nodoc:
        @object = nil
        @key = key
        @store = store
        @token = token
        @options = options
      end

      # Return object if it exists in store, otherwise nil.
      def get
        object
      end

      # Save object in store and set associated token on client.
      def set(object)
        @object = object
      end

      # Remove object from store and delete associated token on client.
      def delete
        set false
      end

      #
      # Create, update, or delete the state in the store and set or delete the 
      # client's cookie accordingly.
      #
      # Note: middleware method not intended for application use.
      #
      # [resp] Rack::Response used to set and delete the client cookie.
      #
      def finish(resp) #:nodoc:
        if token? && object?
          update_state(resp)
        elsif object?
          create_state(resp)
        elsif token? && deleted?
          delete_state(resp)
        end
      end

      private

      attr_reader :store, :key, :options

      def token(generate = false)
        @token = generate ? SecureRandom.urlsafe_base64 : @token
      end

      def token?
        ! @token.nil?
      end

      # Load object from store on initial access only.
      def object
        @object ||= token? ? store.read(token) : nil
      end

      # Returns true if object has been set directly or loaded from store.
      def object?
        !!@object
      end

      # The object has been marked for deletion.
      def deleted?
        @object == false
      end

      def create_state(resp)
        store.create(token(true), object)
        resp.set_cookie(key, options.merge(value: token))
      rescue KeyError # token exists
        retry
      end

      def update_state(resp)
        store.update(token, object)
        resp.set_cookie(key, options.merge(value: token)) # update timestamp
      rescue KeyError # token nonexistent
        create_state(resp)
      end

      def delete_state(resp)
        store.delete(token)
        resp.delete_cookie(key, options)
      end
    end

    #
    # Storage adapters are used by Manager for storing object state.
    #
    # == Available Adapters
    #
    # * File
    # * Memory
    # * Postgres
    # * PreparedPostgres
    #
    # == Adapters Must Implement CRUD Methods
    #
    # [create(token, object)]
    #   Create +object+ at +token+. Raise KeyError if token exists.
    #
    # [read(token)]
    #   Read +object+ at +token+. Return nil if token is nonexistent.
    #
    # [update(token, object)]
    #   Update +object+ at +token+. Raise KeyError if token is nonexistent.
    #
    # [delete(token)]
    #   Delete +object+ at +token+. Return nil if token is nonexistent.
    #
    # == Adapters Must Conform to Specification
    #
    # All adapters must pass the specifications in spec/state_store.rb to
    # guarantee compatibility with Manager.
    #
    module Store

      #
      # The default state storage adapter: a slight subclassing of Hash.
      # Memory is only suitable for development and single process apps.
      #
      class Memory < Hash
        def create(token, object)
          raise KeyError if has_key? token
          store(token, object)
        end

        def read(token)
          fetch(token, nil)
        end

        def update(token, object)
          raise KeyError unless has_key? token
          store(token, object)
        end
      end

      #
      # File-based state storage adapter. The state token is the file name 
      # and the state object is Marshaled and stored in the file.
      #
      class File

        require 'tmpdir'

        #
        # [directory]
        #   Storage directory for state files. It must exist and have proper
        #   permissions.
        #
        def initialize(directory = Dir.tmpdir)
          @directory = directory
        end

        def create(token, object)
          flags = ::File::WRONLY|::File::CREAT|::File::EXCL

          ::File.open(state_file(token), flags, 0600) do |f|
            f.flock ::File::LOCK_EX
            f.write ::Marshal.dump(object)
          end
        rescue Errno::EEXIST
          raise KeyError
        end

        def read(token)
          ::File.open(state_file(token)) do |f|
            f.flock ::File::LOCK_SH
            ::Marshal.load(f) if f.size > 0
          end rescue nil
        end

        def update(token, object)
          flags = ::File::WRONLY|::File::TRUNC

          ::File.open(state_file(token), flags) do |f|
            f.flock ::File::LOCK_EX
            f.write ::Marshal.dump(object)
          end
        rescue Errno::ENOENT
          raise KeyError
        end

        def delete(token)
          ::File.unlink state_file(token) rescue nil
        end

        private

        def state_file(token)
          ::File.join(@directory, token)
        end
      end

      #
      # Postgres table-based state storage adapter.
      #
      # The following statement will create the required database table:
      #
      #   CREATE UNLOGGED TABLE state (
      #     token   varchar PRIMARY KEY,
      #     object  bytea NOT NULL,
      #     mtime   timestamp DEFAULT current_timestamp
      #   )
      #
      # The _mtime_ timestamp column may be used for implementing state expiry
      # external to this class. For example, the following statement could be 
      # piped to psql(1) daily via cron(8):
      #
      #   DELETE FROM state WHERE (mtime + interval '30d') < current_timestamp
      #
      class Postgres

        # Indicate binary data; the object is stored in binary format.
        BINARY = 1

        #
        # [connection]
        #   Instance of the database connection, e.g., PG::Connection.new
        # [table]
        #   Table name where state is stored.
        #
        def initialize(connection, table = 'state')
          @db = connection
          @table = table
        end

        def create(token, object)
          sql = "INSERT INTO #{@table} (token,object) VALUES ($1,$2)"
          params = [token, {value: ::Marshal.dump(object), format: BINARY}]
          @db.exec_params(sql, params).clear
        rescue PG::UniqueViolation
          raise KeyError
        end

        def read(token)
          sql = "SELECT object FROM #{@table} WHERE token = $1"
          @db.exec_params(sql, [token], BINARY) do |result|
            ::Marshal.load result.getvalue(0,0)
          end rescue nil
        end

        def update(token, object)
          sql = "UPDATE #{@table} SET object = $2 WHERE token = $1"
          params = [token, {value: ::Marshal.dump(object), format: BINARY}]
          @db.exec_params(sql, params) do |result|
            raise KeyError if result.cmd_tuples == 0
          end
        end

        def delete(token)
          sql = "DELETE FROM #{@table} WHERE token = $1"
          @db.exec_params(sql, [token]).clear
        end
      end

      #
      # Postgres table-based state storage adapter using prepared statements.
      # Because accessing state is a highly repetitive activity, prepared
      # statements may offer some optimization.
      #
      # See Postgres for details.
      #
      class PreparedPostgres
        BINARY = 1

        def initialize(connection, table = 'state')
          @db = connection
          begin
            @db.prepare 'create_state',
              "INSERT INTO #{table} (token,object) VALUES ($1,$2)"
            @db.prepare 'read_state',
              "SELECT object FROM #{table} WHERE token=$1"
            @db.prepare 'update_state',
              "UPDATE #{table} SET object=$2, mtime=NOW() WHERE token=$1"
            @db.prepare 'delete_state',
              "DELETE FROM #{table} WHERE token=$1"
          rescue PG::DuplicatePstatement
          end
        end

        def create(token, object)
          @db.exec_prepared('create_state', params(token, object)).clear
        rescue PG::UniqueViolation
          raise KeyError
        end

        def read(token)
          @db.exec_prepared('read_state', params(token), BINARY) do |result|
            ::Marshal.load result.getvalue(0,0)
          end rescue nil
        end

        def update(token, object)
          @db.exec_prepared('update_state', params(token, object)) do |result|
            raise KeyError if result.cmd_tuples == 0
          end
        end

        def delete(token)
          @db.exec_prepared('delete_state', params(token)).clear
        end

        private

        def params(token, obj = nil)
          obj ? [token, {value: ::Marshal.dump(obj), format:BINARY}] : [token]
        end
      end
    end
  end
end
